@@ -1,10 +1,58 @@
 //! This crate implements a structure that can be used as a generic array type.
 //!
-//! **Requires minumum Rust version of 1.65.0
+//! **Requires minimum Rust version of 1.65.0
 //!
 //! [Documentation on GH Pages](https://fizyk20.github.io/generic-array/generic_array/)
 //! may be required to view certain types on foreign crates.
 //!
+//! ## Upgrading from 0.14 or using with `hybrid-array 0.4`
+//!
+//! `generic-array 0.14` has been officially deprecated, so here's a quick guide on how to upgrade from `generic-array 0.14` to `1.x`. Note that libraries depending on `generic-array` will need to update their usage as well. Some libraries are moving to `hybrid-array 0.4` instead, which we provide interoperability with `generic-array 1.x` via the `hybrid-array-0_4` feature flag.
+//!
+//! <details>
+//! <summary>Click to expand</summary>
+//!
+//! To upgrade to `1.x`, change your `Cargo.toml` to use the new version:
+//!
+//! ```toml
+//! [dependencies]
+//! generic-array = "1"
+//! ```
+//!
+//! then in your code, go through and remove the `<T>` from `ArrayLength<T>` bounds, as the type parameter has been removed. It's now just `ArrayLength`.
+//!
+//! If you _need_ to interoperate with `generic-array 0.14`, enable the `compat-0_14` feature flag:
+//!
+//! ```toml
+//! [dependencies]
+//! generic-array = { version = "1", features = ["compat-0_14"] }
+//! ```
+//!
+//! then use the `to_0_14`/`from_0_14`/`as_0_14`/`as_0_14_mut` methods on `GenericArray` to convert between versions, or use the `From`/`AsRef`/`AsMut` implementations.
+//!
+//! The `arr!` macro has changed to no longer require a type parameter, so change:
+//!
+//! ```rust,ignore
+//! let array = arr![i32; 1, 2, 3];
+//! // to
+//! let array = arr![1, 2, 3];
+//! ```
+//!
+//! For interoperability with `hybrid-array 0.4`, enable the `hybrid-array-0_4` feature flag:
+//!
+//! ```toml
+//! [dependencies]
+//! generic-array = { version = "1", features = ["hybrid-array-0_4"] }
+//! ```
+//!
+//! then use the `to_ha0_4`/`from_ha0_4`/`as_ha0_4`/`as_ha0_4_mut` methods on `GenericArray` to convert between versions, or use the `From`/`AsRef`/`AsMut` implementations.
+//!
+//! We also implement the `AssocArraySize` and `AsArrayRef`/`AsArrayMut` traits from `hybrid-array` for `GenericArray`.
+//!
+//! </details>
+//!
+//! ## Usage
+//!
 //! Before Rust 1.51, arrays `[T; N]` were problematic in that they couldn't be
 //! generic with respect to the length `N`, so this wouldn't work:
 //!
@@ -80,19 +128,20 @@
 //! ```toml
 //! [dependencies.generic-array]
 //! features = [
-//!     "more_lengths",  # Expands From/Into implementation for more array lengths
-//!     "serde",         # Serialize/Deserialize implementation
-//!     "zeroize",       # Zeroize implementation for setting array elements to zero
-//!     "const-default", # Compile-time const default value support via trait
-//!     "alloc",         # Enables From/TryFrom implementations between GenericArray and Vec<T>/Box<[T]>
-//!     "faster-hex"     # Enables internal use of the `faster-hex` crate for faster hex encoding via SIMD
+//!     "serde",            # Serialize/Deserialize implementation
+//!     "zeroize",          # Zeroize implementation for setting array elements to zero
+//!     "const-default",    # Compile-time const default value support via trait
+//!     "alloc",            # Enables From/TryFrom implementations between GenericArray and Vec<T>/Box<[T]>
+//!     "faster-hex",       # Enables internal use of the `faster-hex` crate for faster hex encoding via SIMD
+//!     "compat-0_14",      # Enables interoperability with `generic-array` 0.14
+//!     "hybrid-array-0_4"  # Enables interoperability with `hybrid-array` 0.4
 //! ]
 //! ```
 
 #![deny(missing_docs)]
 #![deny(meta_variable_misuse)]
 #![no_std]
-#![cfg_attr(docsrs, feature(doc_auto_cfg))]
+#![cfg_attr(docsrs, feature(doc_cfg))]
 
 pub extern crate typenum;
 
@@ -100,6 +149,7 @@ pub extern crate typenum;
 #[cfg(feature = "alloc")]
 pub extern crate alloc;
 
+mod compat;
 mod hex;
 mod impls;
 mod iter;
@@ -133,7 +183,6 @@ pub mod functional;
 pub mod sequence;
 
 mod internal;
-use internal::{ArrayConsumer, IntrusiveArrayBuilder, Sealed};
 
 // re-export to allow doc_auto_cfg to handle it
 #[cfg(feature = "internals")]
@@ -143,19 +192,23 @@ pub mod internals {
     //! These are used internally for building and consuming generic arrays. When used correctly,
     //! they can ensure elements are correctly dropped if something panics while using them.
     //!
-    //! The API of these is not guarenteed to be stable, as they are not intended for general use.
+    //! The API of these is not guaranteed to be stable, as they are not intended for general use.
 
-    pub use crate::internal::{ArrayBuilder, ArrayConsumer, IntrusiveArrayBuilder};
+    pub use crate::internal::{IntrusiveArrayBuilder, IntrusiveArrayConsumer};
+
+    // soft-deprecated
+    pub use crate::internal::{ArrayBuilder, ArrayConsumer};
 }
 
+use internal::{IntrusiveArrayBuilder, IntrusiveArrayConsumer, Sealed};
+
 use self::functional::*;
 use self::sequence::*;
 
 pub use self::iter::GenericArrayIter;
 
-/// Trait used to define the number of elements in a [`GenericArray`].
-///
-/// `ArrayLength` is a superset of [`typenum::Unsigned`].
+/// `ArrayLength` is a type-level [`Unsigned`] integer used to
+/// define the number of elements in a [`GenericArray`].
 ///
 /// Consider `N: ArrayLength` to be equivalent to `const N: usize`
 ///
@@ -176,6 +229,20 @@ pub use self::iter::GenericArrayIter;
 ///
 /// This trait is effectively sealed due to only being allowed on [`Unsigned`] types,
 /// and therefore cannot be implemented in user code.
+///
+/// Furthermore, this is limited to lengths less than or equal to `usize::MAX`.
+/// ```compile_fail
+/// # #![recursion_limit = "256"]
+/// # use generic_array::{GenericArray, ArrayLength};
+/// # use generic_array::typenum::{self, Unsigned};
+/// type Empty = core::convert::Infallible; // Uninhabited ZST, size_of::<Empty>() == 0
+///
+/// // 2^64, greater than usize::MAX on 64-bit systems
+/// type TooBig = typenum::operator_aliases::Shleft<typenum::U1, typenum::U64>;
+///
+/// // Compile Error due to ArrayLength not implemented for TooBig
+/// let _ = GenericArray::<Empty, TooBig>::from_slice(&[]);
+/// ```
 pub unsafe trait ArrayLength: Unsigned + 'static {
     /// Associated type representing the underlying contiguous memory
     /// that constitutes an array with the given number of elements.
@@ -263,15 +330,25 @@ where
 /// Associated [`ArrayLength`] for one [`Const<N>`]
 ///
 /// See [`IntoArrayLength`] for more information.
+///
+/// Note that not all `N` values are valid due to limitations inherent to `typenum` and Rust. You
+/// may need to combine [Const] with other typenum operations to get the desired length.
 pub type ConstArrayLength<const N: usize> = <Const<N> as IntoArrayLength>::ArrayLength;
 
+/// [`GenericArray`] with a const-generic `usize` length, using the [`ConstArrayLength`] type alias for `N`.
+///
+/// To construct from a literal array, use [`from_array`](GenericArray::from_array).
+///
+/// Note that not all `N` values are valid due to limitations inherent to `typenum` and Rust. You
+/// may need to combine [Const] with other typenum operations to get the desired length.
+pub type ConstGenericArray<T, const N: usize> = GenericArray<T, ConstArrayLength<N>>;
+
 /// Internal type used to generate a struct of appropriate size
 #[allow(dead_code)]
 #[repr(C)]
 #[doc(hidden)]
 pub struct GenericArrayImplEven<T, U> {
-    parent1: U,
-    parent2: U,
+    parents: [U; 2],
     _marker: PhantomData<T>,
 }
 
@@ -280,8 +357,7 @@ pub struct GenericArrayImplEven<T, U> {
 #[repr(C)]
 #[doc(hidden)]
 pub struct GenericArrayImplOdd<T, U> {
-    parent1: U,
-    parent2: U,
+    parents: [U; 2],
     data: T,
 }
 
@@ -309,12 +385,32 @@ impl<T: Copy, U: Copy> Copy for GenericArrayImplOdd<T, U> {}
 impl<T, U> Sealed for GenericArrayImplEven<T, U> {}
 impl<T, U> Sealed for GenericArrayImplOdd<T, U> {}
 
-unsafe impl<N: ArrayLength> ArrayLength for UInt<N, B0> {
+// 1 << (size_of::<usize>() << 3) == usize::MAX + 1
+type MaxArrayLengthP1 = typenum::Shleft<
+    typenum::U1,
+    typenum::Shleft<typenum::U<{ mem::size_of::<usize>() }>, typenum::U3>,
+>;
+
+/// Helper trait to hide the complex bound under a simpler name
+trait IsWithinUsizeBound: typenum::IsLess<MaxArrayLengthP1, Output = typenum::consts::True> {}
+
+impl<N> IsWithinUsizeBound for N where
+    N: typenum::IsLess<MaxArrayLengthP1, Output = typenum::consts::True>
+{
+}
+
+unsafe impl<N: ArrayLength> ArrayLength for UInt<N, B0>
+where
+    Self: IsWithinUsizeBound,
+{
     #[doc(hidden)]
     type ArrayType<T> = GenericArrayImplEven<T, N::ArrayType<T>>;
 }
 
-unsafe impl<N: ArrayLength> ArrayLength for UInt<N, B1> {
+unsafe impl<N: ArrayLength> ArrayLength for UInt<N, B1>
+where
+    Self: IsWithinUsizeBound,
+{
     #[doc(hidden)]
     type ArrayType<T> = GenericArrayImplOdd<T, N::ArrayType<T>>;
 }
@@ -326,7 +422,7 @@ unsafe impl<N: ArrayLength> ArrayLength for UInt<N, B1> {
 ///
 /// # Usage Notes
 ///
-/// ### Intialization
+/// ### Initialization
 ///
 /// Initialization of known-length `GenericArray`s can be done via the [`arr![]`](arr!) macro,
 /// or [`from_array`](GenericArray::from_array)/[`from_slice`](GenericArray::from_slice).
@@ -512,8 +608,8 @@ where
         F: FnMut(usize) -> T,
     {
         unsafe {
-            let mut array = GenericArray::<T, N>::uninit();
-            let mut builder = IntrusiveArrayBuilder::new(&mut array);
+            let mut array = MaybeUninit::<GenericArray<T, N>>::uninit();
+            let mut builder = IntrusiveArrayBuilder::new_alt(&mut array);
 
             {
                 let (builder_iter, position) = builder.iter_position();
@@ -524,8 +620,7 @@ where
                 });
             }
 
-            builder.finish();
-            IntrusiveArrayBuilder::array_assume_init(array)
+            builder.finish_and_assume_init()
         }
     }
 
@@ -543,8 +638,11 @@ where
     {
         unsafe {
             if mem::needs_drop::<T>() || mem::needs_drop::<B>() {
-                let mut left = ArrayConsumer::new(lhs);
-                let mut right = ArrayConsumer::new(self);
+                let mut left = ManuallyDrop::new(lhs);
+                let mut right = ManuallyDrop::new(self);
+
+                let mut left = IntrusiveArrayConsumer::new(&mut left);
+                let mut right = IntrusiveArrayConsumer::new(&mut right);
 
                 let (left_array_iter, left_position) = left.iter_position();
                 let (right_array_iter, right_position) = right.iter_position();
@@ -584,7 +682,8 @@ where
     {
         unsafe {
             if mem::needs_drop::<T>() {
-                let mut right = ArrayConsumer::new(self);
+                let mut right = ManuallyDrop::new(self);
+                let mut right = IntrusiveArrayConsumer::new(&mut right);
 
                 let (right_array_iter, right_position) = right.iter_position();
 
@@ -625,7 +724,8 @@ where
         F: FnMut(T) -> U,
     {
         unsafe {
-            let mut source = ArrayConsumer::new(self);
+            let mut array = ManuallyDrop::new(self);
+            let mut source = IntrusiveArrayConsumer::new(&mut array);
 
             let (array_iter, position) = source.iter_position();
 
@@ -656,7 +756,8 @@ where
         F: FnMut(U, T) -> U,
     {
         unsafe {
-            let mut source = ArrayConsumer::new(self);
+            let mut array = ManuallyDrop::new(self);
+            let mut source = IntrusiveArrayConsumer::new(&mut array);
 
             let (array_iter, position) = source.iter_position();
 
@@ -686,6 +787,9 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
     }
 
     /// Extracts a mutable slice containing the entire array.
+    ///
+    /// This method is `const` since Rust 1.83.0, but non-`const` before.
+    #[rustversion::attr(since(1.83), const)]
     #[inline(always)]
     pub fn as_mut_slice(&mut self) -> &mut [T] {
         unsafe { slice::from_raw_parts_mut(self as *mut Self as *mut T, N::USIZE) }
@@ -728,11 +832,13 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
     /// Panics if the slice is not equal to the length of the array.
     ///
     /// Consider [`TryFrom`]/[`TryInto`] for a fallible conversion.
+    ///
+    /// This method is `const` since Rust 1.83.0, but non-`const` before.
+    #[rustversion::attr(since(1.83), const)]
     #[inline(always)]
     pub fn from_mut_slice(slice: &mut [T]) -> &mut GenericArray<T, N> {
-        assert_eq!(
-            slice.len(),
-            N::USIZE,
+        assert!(
+            slice.len() == N::USIZE,
             "slice.len() != N in GenericArray::from_mut_slice"
         );
 
@@ -742,11 +848,16 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
     /// Converts a mutable slice to a mutable generic array reference with inferred length.
     ///
     /// This is a fallible alternative to [`from_mut_slice`](GenericArray::from_mut_slice),
-    /// and current just calls [`TryFrom`] internally, but is provided for
-    /// future compatibility when we can make it const.
+    /// and is equivalent to the [`TryFrom`] implementation with the added benefit of being `const`.
+    ///
+    /// This method is `const` since Rust 1.83.0, but non-`const` before.
+    #[rustversion::attr(since(1.83), const)]
     #[inline(always)]
     pub fn try_from_mut_slice(slice: &mut [T]) -> Result<&mut GenericArray<T, N>, LengthError> {
-        TryFrom::try_from(slice)
+        match slice.len() == N::USIZE {
+            true => Ok(GenericArray::from_mut_slice(slice)),
+            false => Err(LengthError),
+        }
     }
 
     /// Converts a slice of `T` elements into a slice of `GenericArray<T, N>` chunks.
@@ -782,6 +893,9 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
     /// # Panics
     ///
     /// Panics if `N` is `U0` _AND_ the input slice is not empty.
+    ///
+    /// This method is `const` since Rust 1.83.0, but non-`const` before.
+    #[rustversion::attr(since(1.83), const)]
     pub fn chunks_from_slice_mut(slice: &mut [T]) -> (&mut [GenericArray<T, N>], &mut [T]) {
         if N::USIZE == 0 {
             assert!(slice.is_empty(), "GenericArray length N must be non-zero");
@@ -811,6 +925,9 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
     }
 
     /// Convert a slice of `GenericArray<T, N>` into a slice of `T`, effectively flattening the arrays.
+    ///
+    /// This method is `const` since Rust 1.83.0, but non-`const` before.
+    #[rustversion::attr(since(1.83), const)]
     #[inline(always)]
     pub fn slice_from_chunks_mut(slice: &mut [GenericArray<T, N>]) -> &mut [T] {
         unsafe { slice::from_raw_parts_mut(slice.as_mut_ptr() as *mut T, slice.len() * N::USIZE) }
@@ -848,6 +965,9 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
     }
 
     /// Convert a mutable slice of native arrays into a mutable slice of `GenericArray`s.
+    ///
+    /// This method is `const` since Rust 1.83.0, but non-`const` before.
+    #[rustversion::attr(since(1.83), const)]
     #[inline(always)]
     pub fn from_chunks_mut<const U: usize>(chunks: &mut [[T; U]]) -> &mut [GenericArray<T, N>]
     where
@@ -866,6 +986,9 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
     }
 
     /// Converts a mutable slice `GenericArray<T, N>` into a mutable slice of `[T; N]`
+    ///
+    /// This method is `const` since Rust 1.83.0, but non-`const` before.
+    #[rustversion::attr(since(1.83), const)]
     #[inline(always)]
     pub fn into_chunks_mut<const U: usize>(chunks: &mut [GenericArray<T, N>]) -> &mut [[T; U]]
     where
@@ -913,7 +1036,7 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
     /// ```
     #[inline(always)]
     pub const unsafe fn assume_init(array: GenericArray<MaybeUninit<T>, N>) -> Self {
-        const_transmute::<_, MaybeUninit<GenericArray<T, N>>>(array).assume_init()
+        const_transmute::<GenericArray<MaybeUninit<T>, N>, GenericArray<T, N>>(array)
     }
 }
 
@@ -921,7 +1044,8 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
 #[derive(Debug, Clone, Copy)]
 pub struct LengthError;
 
-// TODO: Impl core::error::Error when when https://github.com/rust-lang/rust/issues/103765 is finished
+#[rustversion::since(1.81)]
+impl core::error::Error for LengthError {}
 
 impl core::fmt::Display for LengthError {
     fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
@@ -943,10 +1067,7 @@ impl<'a, T, N: ArrayLength> TryFrom<&'a mut [T]> for &'a mut GenericArray<T, N>
 
     #[inline(always)]
     fn try_from(slice: &'a mut [T]) -> Result<Self, Self::Error> {
-        match slice.len() == N::USIZE {
-            true => Ok(GenericArray::from_mut_slice(slice)),
-            false => Err(LengthError),
-        }
+        GenericArray::try_from_mut_slice(slice)
     }
 }
 
@@ -972,8 +1093,8 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
         }
 
         unsafe {
-            let mut array = GenericArray::uninit();
-            let mut builder = IntrusiveArrayBuilder::new(&mut array);
+            let mut array = MaybeUninit::<GenericArray<T, N>>::uninit();
+            let mut builder = IntrusiveArrayBuilder::new_alt(&mut array);
 
             builder.extend(&mut iter);
 
@@ -981,34 +1102,58 @@ impl<T, N: ArrayLength> GenericArray<T, N> {
                 return Err(LengthError);
             }
 
-            Ok({
-                builder.finish();
-                IntrusiveArrayBuilder::array_assume_init(array)
-            })
+            Ok(builder.finish_and_assume_init())
         }
     }
 }
 
 /// A const reimplementation of the [`transmute`](core::mem::transmute) function,
-/// avoiding problems when the compiler can't prove equal sizes.
+/// avoiding problems when the compiler can't prove equal sizes for some reason.
+///
+/// This will still check that the sizes of `A` and `B` are equal at compile time:
+/// ```compile_fail
+/// # use generic_array::const_transmute;
+///
+/// let _ = unsafe { const_transmute::<u32, u64>(0u32) }; // panics at compile time
+/// ```
 ///
 /// # Safety
 /// Treat this the same as [`transmute`](core::mem::transmute), or (preferably) don't use it at all.
 #[inline(always)]
 #[cfg_attr(not(feature = "internals"), doc(hidden))]
 pub const unsafe fn const_transmute<A, B>(a: A) -> B {
-    if mem::size_of::<A>() != mem::size_of::<B>() {
-        panic!("Size mismatch for generic_array::const_transmute");
+    struct SizeAsserter<A, B>(PhantomData<(A, B)>);
+
+    impl<A, B> SizeAsserter<A, B> {
+        const ASSERT_SIZE_EQUALITY: () = {
+            if mem::size_of::<A>() != mem::size_of::<B>() {
+                panic!("Size mismatch for generic_array::const_transmute");
+            }
+        };
     }
 
-    #[repr(C)]
-    union Union<A, B> {
-        a: ManuallyDrop<A>,
-        b: ManuallyDrop<B>,
+    let () = SizeAsserter::<A, B>::ASSERT_SIZE_EQUALITY;
+
+    #[rustversion::since(1.83)]
+    #[inline(always)]
+    const unsafe fn do_transmute<A, B>(a: A) -> B {
+        mem::transmute_copy(&ManuallyDrop::new(a))
     }
 
-    let a = ManuallyDrop::new(a);
-    ManuallyDrop::into_inner(Union { a }.b)
+    #[rustversion::before(1.83)]
+    #[inline(always)]
+    const unsafe fn do_transmute<A, B>(a: A) -> B {
+        #[repr(C)]
+        union Union<A, B> {
+            a: ManuallyDrop<A>,
+            b: ManuallyDrop<B>,
+        }
+
+        let a = ManuallyDrop::new(a);
+        ManuallyDrop::into_inner(Union { a }.b)
+    }
+
+    do_transmute(a)
 }
 
 #[cfg(test)]
